@@ -5,10 +5,18 @@ from math import factorial, log
 from sklearn.neighbors import KDTree
 from scipy.signal import periodogram, welch
 
-from .utils import _embed
+from .utils import _embed, _xlogx
 
-all = ['perm_entropy', 'spectral_entropy', 'svd_entropy', 'app_entropy',
-       'sample_entropy', 'lziv_complexity', 'num_zerocross', 'hjorth_params']
+all = [
+    "perm_entropy",
+    "spectral_entropy",
+    "svd_entropy",
+    "app_entropy",
+    "sample_entropy",
+    "lziv_complexity",
+    "num_zerocross",
+    "hjorth_params",
+]
 
 
 def perm_entropy(x, order=3, delay=1, normalize=False):
@@ -20,8 +28,10 @@ def perm_entropy(x, order=3, delay=1, normalize=False):
         One-dimensional time series of shape (n_times)
     order : int
         Order of permutation entropy. Default is 3.
-    delay : int
-        Time delay (lag). Default is 1.
+    delay : int, list, np.ndarray or range
+        Time delay (lag). Default is 1. If multiple values are passed
+        (e.g. [1, 2, 3]), AntroPy will calculate the average permutation
+        entropy across all these delays.
     normalize : bool
         If True, divide by log2(order!) to normalize the entropy between 0
         and 1. Otherwise, return the permutation entropy in bit.
@@ -80,26 +90,18 @@ def perm_entropy(x, order=3, delay=1, normalize=False):
     >>> print(f"{ant.perm_entropy(x, normalize=True):.4f}")
     0.5888
 
-    Fractional Gaussian noise with H = 0.5
-
+    Fractional Gaussian noise with H = 0.5, averaged across multiple delays
     >>> rng = np.random.default_rng(seed=42)
     >>> x = sn.FractionalGaussianNoise(hurst=0.5, rng=rng).sample(10000)
-    >>> print(f"{ant.perm_entropy(x, normalize=True):.4f}")
-    0.9998
+    >>> print(f"{ant.perm_entropy(x, delay=[1, 2, 3], normalize=True):.4f}")
+    0.9999
 
-    Fractional Gaussian noise with H = 0.9
-
-    >>> rng = np.random.default_rng(seed=42)
-    >>> x = sn.FractionalGaussianNoise(hurst=0.9, rng=rng).sample(10000)
-    >>> print(f"{ant.perm_entropy(x, normalize=True):.4f}")
-    0.9926
-
-    Fractional Gaussian noise with H = 0.1
+    Fractional Gaussian noise with H = 0.1, averaged across multiple delays
 
     >>> rng = np.random.default_rng(seed=42)
     >>> x = sn.FractionalGaussianNoise(hurst=0.1, rng=rng).sample(10000)
-    >>> print(f"{ant.perm_entropy(x, normalize=True):.4f}")
-    0.9959
+    >>> print(f"{ant.perm_entropy(x, delay=[1, 2, 3], normalize=True):.4f}")
+    0.9986
 
     Random
 
@@ -119,25 +121,28 @@ def perm_entropy(x, order=3, delay=1, normalize=False):
     >>> print(f"{ant.perm_entropy(x, normalize=True):.4f}")
     -0.0000
     """
+    # If multiple delay are passed, return the average across all d
+    if isinstance(delay, (list, np.ndarray, range)):
+        return np.mean([perm_entropy(x, order=order, delay=d, normalize=normalize) for d in delay])
     x = np.array(x)
     ran_order = range(order)
     hashmult = np.power(order, ran_order)
+    assert delay > 0, "delay must be greater than zero."
     # Embed x and sort the order of permutations
-    sorted_idx = _embed(x, order=order, delay=delay).argsort(kind='quicksort')
+    sorted_idx = _embed(x, order=order, delay=delay).argsort(kind="quicksort")
     # Associate unique integer to each permutations
     hashval = (np.multiply(sorted_idx, hashmult)).sum(1)
     # Return the counts
     _, c = np.unique(hashval, return_counts=True)
     # Use np.true_divide for Python 2 compatibility
     p = np.true_divide(c, c.sum())
-    pe = -np.multiply(p, np.log2(p)).sum()
+    pe = -_xlogx(p).sum()
     if normalize:
         pe /= np.log2(factorial(order))
     return pe
 
 
-def spectral_entropy(x, sf, method='fft', nperseg=None, normalize=False,
-                     axis=-1):
+def spectral_entropy(x, sf, method="fft", nperseg=None, normalize=False, axis=-1):
     """Spectral Entropy.
 
     Parameters
@@ -203,12 +208,12 @@ def spectral_entropy(x, sf, method='fft', nperseg=None, normalize=False,
     >>> np.random.seed(42)
     >>> x = np.random.rand(3000)
     >>> ant.spectral_entropy(x, sf=100, method='welch')
-    6.980045662371389
+    6.98004566237139
 
     Normalized spectral entropy
 
     >>> ant.spectral_entropy(x, sf=100, method='welch', normalize=True)
-    0.9955526198316071
+    0.9955526198316073
 
     Normalized spectral entropy of 2D data
 
@@ -241,12 +246,12 @@ def spectral_entropy(x, sf, method='fft', nperseg=None, normalize=False,
     """
     x = np.asarray(x)
     # Compute and normalize power spectrum
-    if method == 'fft':
+    if method == "fft":
         _, psd = periodogram(x, sf, axis=axis)
-    elif method == 'welch':
+    elif method == "welch":
         _, psd = welch(x, sf, nperseg=nperseg, axis=axis)
     psd_norm = psd / psd.sum(axis=axis, keepdims=True)
-    se = -(psd_norm * np.log2(psd_norm)).sum(axis=axis)
+    se = -_xlogx(psd_norm).sum(axis=axis)
     if normalize:
         se /= np.log2(psd_norm.shape[axis])
     return se
@@ -359,19 +364,20 @@ def svd_entropy(x, order=3, delay=1, normalize=False):
     W = np.linalg.svd(mat, compute_uv=False)
     # Normalize the singular values
     W /= sum(W)
-    svd_e = -np.multiply(W, np.log2(W)).sum()
+    svd_e = -_xlogx(W).sum()
     if normalize:
         svd_e /= np.log2(order)
     return svd_e
 
 
-def _app_samp_entropy(x, order, metric='chebyshev', approximate=True):
-    """Utility function for `app_entropy`` and `sample_entropy`.
-    """
+def _app_samp_entropy(x, order, metric="chebyshev", approximate=True):
+    """Utility function for `app_entropy`` and `sample_entropy`."""
     _all_metrics = KDTree.valid_metrics
     if metric not in _all_metrics:
-        raise ValueError('The given metric (%s) is not valid. The valid '
-                         'metric names are: %s' % (metric, _all_metrics))
+        raise ValueError(
+            "The given metric (%s) is not valid. The valid "
+            "metric names are: %s" % (metric, _all_metrics)
+        )
     phi = np.zeros(2)
     r = 0.2 * np.std(x, ddof=0)
 
@@ -381,14 +387,18 @@ def _app_samp_entropy(x, order, metric='chebyshev', approximate=True):
         emb_data1 = _emb_data1
     else:
         emb_data1 = _emb_data1[:-1]
-    count1 = KDTree(emb_data1, metric=metric).query_radius(emb_data1, r,
-                                                           count_only=True
-                                                           ).astype(np.float64)
+    count1 = (
+        KDTree(emb_data1, metric=metric)
+        .query_radius(emb_data1, r, count_only=True)
+        .astype(np.float64)
+    )
     # compute phi(order + 1, r)
     emb_data2 = _embed(x, order + 1, 1)
-    count2 = KDTree(emb_data2, metric=metric).query_radius(emb_data2, r,
-                                                           count_only=True
-                                                           ).astype(np.float64)
+    count2 = (
+        KDTree(emb_data2, metric=metric)
+        .query_radius(emb_data2, r, count_only=True)
+        .astype(np.float64)
+    )
     if approximate:
         phi[0] = np.mean(np.log(count1 / emb_data1.shape[0]))
         phi[1] = np.mean(np.log(count2 / emb_data2.shape[0]))
@@ -398,7 +408,7 @@ def _app_samp_entropy(x, order, metric='chebyshev', approximate=True):
     return phi
 
 
-@jit('f8(f8[:], i4, f8)', nopython=True)
+@jit("f8(f8[:], i4, f8)", nopython=True)
 def _numba_sampen(x, order, r):
     """
     Fast evaluation of the sample entropy using Numba.
@@ -453,7 +463,7 @@ def _numba_sampen(x, order, r):
     return -log(p[-1])
 
 
-def app_entropy(x, order=2, metric='chebyshev'):
+def app_entropy(x, order=2, metric="chebyshev"):
     """Approximate Entropy.
 
     Parameters
@@ -545,7 +555,7 @@ def app_entropy(x, order=2, metric='chebyshev'):
     return np.subtract(phi[0], phi[1])
 
 
-def sample_entropy(x, order=2, metric='chebyshev'):
+def sample_entropy(x, order=2, metric="chebyshev"):
     """Sample Entropy.
 
     Parameters
@@ -651,44 +661,57 @@ def sample_entropy(x, order=2, metric='chebyshev'):
     -0.0000
     """
     x = np.asarray(x, dtype=np.float64)
-    if metric == 'chebyshev' and x.size < 5000:
+    if metric == "chebyshev" and x.size < 5000:
         return _numba_sampen(x, order=order, r=(0.2 * x.std(ddof=0)))
     else:
-        phi = _app_samp_entropy(x, order=order, metric=metric,
-                                approximate=False)
+        phi = _app_samp_entropy(x, order=order, metric=metric, approximate=False)
         return -np.log(np.divide(phi[1], phi[0]))
 
 
-@jit('u8(unicode_type)', nopython=True)
+@jit("uint32(uint32[:])", nopython=True)
 def _lz_complexity(binary_string):
     """Internal Numba implementation of the Lempel-Ziv (LZ) complexity.
-
     https://github.com/Naereen/Lempel-Ziv_Complexity/blob/master/src/lziv_complexity.py
+    - Updated with strict integer typing instead of strings
+    - Slight restructuring based on Yacine Mahdid's notebook:
+    https://github.com/BIAPT/Notebooks/blob/master/features/Lempel-Ziv%20Complexity.ipynb
     """
-    u, v, w = 0, 1, 1
-    v_max = 1
-    length = len(binary_string)
+    # Initialize variables
     complexity = 1
-    while True:
-        if binary_string[u + v - 1] == binary_string[w + v - 1]:
-            v += 1
-            if w + v >= length:
-                complexity += 1
-                break
+    prefix_len = 1
+    len_substring = 1
+    max_len_substring = 1
+    pointer = 0
+
+    # Iterate until the entire string has not been parsed
+    while prefix_len + len_substring <= len(binary_string):
+        # Given a prefix length, find the largest substring
+        if (
+            binary_string[pointer + len_substring - 1]
+            == binary_string[prefix_len + len_substring - 1]
+        ):
+            len_substring += 1  # increase the length of the substring
         else:
-            v_max = max(v, v_max)
-            u += 1
-            if u == w:
+            max_len_substring = max(len_substring, max_len_substring)
+            pointer += 1
+            # Since all pointers have been scanned, pick largest as the jump
+            # size
+            if pointer == prefix_len:
+                # Increment complexity
                 complexity += 1
-                w += v_max
-                if w >= length:
-                    break
-                else:
-                    u = 0
-                    v = 1
-                    v_max = 1
-            else:
-                v = 1
+                # Set prefix length to the max substring size found so far
+                # (jump size)
+                prefix_len += max_len_substring
+                # Reset pointer and max substring size
+                pointer = 0
+                max_len_substring = 1
+            # Reset length of current substring
+            len_substring = 1
+
+    # Check if final iteration occurred in the middle of a substring
+    if len_substring != 1:
+        complexity += 1
+
     return complexity
 
 
@@ -735,14 +758,11 @@ def lziv_complexity(sequence, normalize=False):
       IEEE Transactions on Information Theory / Professional Technical
       Group on Information Theory, 22(1), 75–81.
       https://doi.org/10.1109/TIT.1976.1055501
-
     * Zhang, Y., Hao, J., Zhou, C., & Chang, K. (2009). Normalized
       Lempel-Ziv complexity and its application in bio-sequence analysis.
       Journal of Mathematical Chemistry, 46(4), 1203–1212.
       https://doi.org/10.1007/s10910-008-9512-2
-
     * https://en.wikipedia.org/wiki/Lempel-Ziv_complexity
-
     * https://github.com/Naereen/Lempel-Ziv_Complexity
 
     Examples
@@ -753,18 +773,18 @@ def lziv_complexity(sequence, normalize=False):
     >>> lziv_complexity(s)
     6
 
-    Using a list of integer / boolean instead of a string:
+    Using a list of integer / boolean instead of a string
 
     >>> # 1 / 0 / 10
     >>> lziv_complexity([1, 0, 1, 0, 1, 0, 1, 0, 1, 0])
     3
 
-    With normalization:
+    With normalization
 
     >>> lziv_complexity(s, normalize=True)
     1.5
 
-    This function also works with characters and words:
+    This function also works with characters and words
 
     >>> s = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
     >>> lziv_complexity(s), lziv_complexity(s, normalize=True)
@@ -778,13 +798,17 @@ def lziv_complexity(sequence, normalize=False):
     assert isinstance(normalize, bool)
     if isinstance(sequence, (list, np.ndarray)):
         sequence = np.asarray(sequence)
-        if sequence.dtype.kind in 'bfi':
+        if sequence.dtype.kind in "bfi":
             # Convert [True, False] or [1., 0.] to [1, 0]
-            sequence = sequence.astype(int)
-        # Convert to a string, e.g. "10001100"
-        s = ''.join(sequence.astype(str))
+            s = sequence.astype("uint32")
+        else:
+            # Treat as numpy array of strings
+            # Map string characters to utf-8 integer representation
+            s = np.fromiter(map(ord, "".join(sequence.astype(str))), dtype="uint32")
+            # Can't preallocate length (by specifying count) due to string
+            # concatenation
     else:
-        s = sequence
+        s = np.fromiter(map(ord, sequence), dtype="uint32")
 
     if normalize:
         # 1) Timmermann et al. 2019
@@ -799,7 +823,7 @@ def lziv_complexity(sequence, normalize=False):
         # return _lz_complexity(s) / _lz_complexity(s_shuffled)
         # 2) Zhang et al. 2009
         n = len(s)
-        base = len(''.join(set(s)))  # Number of unique characters
+        base = sum(np.bincount(s) > 0)  # Number of unique characters
         base = 2 if base < 2 else base
         return _lz_complexity(s) / (n / log(n, base))
     else:
@@ -810,6 +834,7 @@ def lziv_complexity(sequence, normalize=False):
 # OTHER TIME-DOMAIN METRICS
 ###############################################################################
 
+
 def num_zerocross(x, normalize=False, axis=-1):
     """Number of zero-crossings.
 
